#ifndef MBS_H
#define MBS_H
#include <stddef.h>

#define CHMASK(c) ((c) & 0xff)
#define RANGE(A, V, B) (((A) <= (V)) && ((V) <= (B)))

#define ISHKANA(c)  RANGE(0xa1, CHMASK(c), 0xdf)

/* sample:
  for (size_t i = 0, b; b = sjisclen(s.c_str(), i, s.size()); i += b) {
    std::cout << "[" << std::string(&(s[i]), b) << "]";
  }
*/

/* SJIS */
#define ISSJIS1H(c)  RANGE(0x81, CHMASK(c), 0x9f)
#define ISSJIS2H(c)  RANGE(0xe0, CHMASK(c), 0xef)
#define ISSJISGH(c)  RANGE(0xf0, CHMASK(c), 0xf9) /* GAIJI */
#define ISSJISIH(c)  RANGE(0xfa, CHMASK(c), 0xfc) /* IBM(SEL) */

#define ISSJISH(c)   (ISSJIS1H(c) || ISSJIS2H(c))
#define ISSJISL(c)   RANGE(0x40, CHMASK(c), 0xfc)

/* UJIS */
#define ISUJIS1H(c)  RANGE(0xa1, CHMASK(c), 0xa8)
#define ISUJIS2H(c)  RANGE(0xb0, CHMASK(c), 0xf4)
#define ISUJISU1H(c) RANGE(0xf5, CHMASK(c), 0xfe) /* USERDEF1 */
#define ISUJISU2H(c) (CHMASK(c) == 0x8f)          /* USERDEF2 */
#define ISUKANA(c)   (CHMASK(c) == 0x8e)          /* SS2 */

#define ISUJISH(c)   (ISUJIS1H(c) || ISUJIS2H(c))
#define ISUJISL(c)   RANGE(0xa1, CHMASK(c), 0xfe)

/* UTF8 */
#define UTF8BITCHECK(c) \
 ((((c) & 0x80) == 0) ? 0 : \
  ((((c) & 0x40) == 0) ? 1 : \
   ((((c) & 0x20) == 0) ? 2 : \
    ((((c) & 0x10) == 0) ? 3 : \
     ((((c) & 0x08) == 0) ? 4 : \
      ((((c) & 0x04) == 0) ? 5 : \
       ((((c) & 0x02) == 0) ? 6 : -1)))))))

/*
return zero if end-of-line
else return mbchar-length
inline size_t utf8clen(const char *buff, size_t i, size_t l);
inline size_t sjisclen(const char *buff, size_t i, size_t l);
inline size_t ujisclen(const char *buff, size_t i, size_t l);
*/

#ifdef MBS_DEF_FUNC
#define inline /* inline is not supported by ANSI-C, but C99 supports. */
#elif defined(MBS_REF_FUNC)
extern size_t utf8clen(const char *buff, size_t i, size_t l);
extern size_t utf8clen(const char *buff, size_t i, size_t l);
extern size_t utf8clen(const char *buff, size_t i, size_t l);
#elif !defined(__cplusplus)
#if defined(_MSC_VER) || defined(__BORLANDC__) || defined(__GNUC__)
#define inline  __inline
#endif
#endif

#ifndef MBS_REF_FUNC
inline size_t utf8clen(const char *buff, size_t i, size_t l) {
  size_t n, j;
  if (i >= l) {
    return 0;
  }
  n = UTF8BITCHECK(buff[i] & 0xff);
  if (n == -1) { /* lost */
    return 1;
  } else if ((i + n) >= l) { /* lost */
    return 1;
  } else if (n == 0) { 
    return 1;
  } else if (n == 1) { /* lost */
    return 1;
  } else {
    for (j = 1; j < n; j++) {
      if (UTF8BITCHECK(buff[i + j] & 0xff) != 1) { /* lost */
        return 1;
      }
    }
    return n;
  }
}

inline size_t sjisclen(const char *buff, size_t i, size_t l) {
  if (i >= l) {
    return 0;
  } else if (i == (l - 1)) {
    return 1;
  } else if (ISSJISH(buff[i + 0]) && ISSJISL(buff[i + 1])) {
    return 2;
  } else {
    return 1;
  }
}

inline size_t ujisclen(const char *buff, size_t i, size_t l) {
  if (i >= l) {
    return 0;
  } else if (i == (l - 1)) {
    return 1;
  } else if (ISUJISH(buff[i + 0]) && ISUJISL(buff[i + 1])) {
    return 2;
  } else if (ISUKANA(buff[i + 0]) && ISHKANA(buff[i + 1])) {
    return 2;
  } else {
    return 1;
  }
}
#endif

#endif
