%include {
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "sx.h"

#define SETBIT(m, n) ((m)[(n) / 8] |= (1 << ((n) % 8)))
#define CLRBIT(m, n) ((m)[(n) / 8] &= ~(1 << ((n) % 8)))
#define CHKBIT(m, n) ((m)[(n) / 8] & (1 << ((n) % 8)))
#define BMLEN(n) ((n) + 7 / 8)
#define NTOKEN(p) (p[2])
#define NSTATE(p) (p[1])
#define ACCNUM(p) (p[1] - 1)
#define SNBUFF(p) (&(p[3 + (NTOKEN(p) * 4)]))
#define FPBUFF(p) (&(p[3 + (NTOKEN(p) * 4) + NSTATE(p)]))
#define FOLLOWPOS(p, s, d) (FPBUFF(p)[(s) * NSTATE(p) + (d)])
#define TV(p, n) ((p)[3 + ((n) * 4) + 0])
#define TS(p, n) ((p)[3 + ((n) * 4) + 1])
#define TL(p, n) ((p)[3 + ((n) * 4) + 2])
#define SN(p, n) ((p)[3 + ((n) * 4) + 3])
#define TN(p, n) (SNBUFF(p)[n])
#define BMSIZE(p) BMLEN(NSTATE(p)) /* in action only */
#define BMADD(p, d, s) \
 { int i; for (i = 0; i < BMSIZE(p); i++) d[i] |= s[i]; }
#define BMCPY(p, d, s) \
  memcpy(d, s, BMSIZE(p))


struct node {
  int na;
  char *fpos;
  char *lpos;
};

struct node *newnode(int n) {
  size_t bs = (size_t)BMLEN(n);
  size_t ns = sizeof (struct node);
  struct node *p = (struct node *)malloc(ns + bs + bs);
  if (p) {
    p->fpos = (char *)&(p[1]);
    p->lpos = &((p->fpos)[bs]);
    memset(p->fpos, 0, bs + bs);
  } else {
    fprintf(stderr, "malloc error\n");
    exit(-1);
  }
  return p;
}

}

%stack_overflow { fprintf(stderr,"Giving up.  Parser stack overflow\n"); }
%name alamo

%type lor { struct node * }
%type seq { struct node * }
%type exp { struct node * }
%type val { struct node * }
%token_type { unsigned short }

%extra_argument { unsigned short *tokens }
/* %token_prefix TK_ */

%right AST PLS QST.
%left LOR.
%nonassoc EOE.

program ::= lor(B) EOE.
  { int i;
    for (i = 0; i < NSTATE(tokens); i++) {
      if (CHKBIT(B->fpos, i)) {
        FOLLOWPOS(tokens, 0, i) = TN(tokens, i);
      }
    }
    for (i = 0; i < NSTATE(tokens); i++) {
      if (CHKBIT(B->lpos, i)) {
        FOLLOWPOS(tokens, ACCNUM(tokens), i) = TN(tokens, i);
      }
    }
    free(B);
  }

lor(A) ::= exp(B) LOR exp(C). 
  { 
    A = newnode(NSTATE(tokens)); 
    A->na = (B->na | C->na);
    BMCPY(tokens, A->fpos, B->fpos);
    BMADD(tokens, A->fpos, C->fpos);
    BMCPY(tokens, A->lpos, B->lpos);
    BMADD(tokens, A->lpos, C->lpos);
    free(B);
    free(C);
  }

lor(A) ::= exp(B).
  { A = B; }

exp(A) ::= exp(B) val(C).
  { int y, x;
    A = newnode(NSTATE(tokens)); 
    A->na = (B->na & C->na);
    BMCPY(tokens, A->fpos, B->fpos);
    if (B->na) { BMADD(tokens, A->fpos, C->fpos); }
    BMCPY(tokens, A->lpos, C->lpos);
    if (C->na) { BMADD(tokens, A->lpos, B->lpos); }
    for (y = 0; y < NSTATE(tokens); y++) {
      if (CHKBIT(B->lpos, y)) {
        for (x = 0; x < NSTATE(tokens); x++) {
          if (CHKBIT(C->fpos, x)) {
            FOLLOWPOS(tokens, y, x) = TN(tokens, x);
          }
        }
      }
    }
    free(B);
    free(C);
  }

exp(A) ::= val(B).
  { A = B; }

val(A) ::= val(B) AST. 
  { int y, x;
    A = newnode(NSTATE(tokens)); 
    A->na = 1;
    BMCPY(tokens, A->fpos, B->fpos);
    BMCPY(tokens, A->lpos, B->lpos);
    for (y = 0; y < NSTATE(tokens); y++) {
      if (CHKBIT(A->lpos, y)) {
        for (x = 0; x < NSTATE(tokens); x++) {
          if (CHKBIT(A->fpos, x)) {
            FOLLOWPOS(tokens, y, x) = TN(tokens, x);
          }
        }
      }
    }
    free(B);
  }

val(A) ::= val(B) PLS. 
  { int y, x;
    A = newnode(NSTATE(tokens)); 
    A->na = 0;
    BMCPY(tokens, A->fpos, B->fpos);
    BMCPY(tokens, A->lpos, B->lpos); 
    for (y = 0; y < NSTATE(tokens); y++) {
      if (CHKBIT(A->lpos, y)) {
        for (x = 0; x < NSTATE(tokens); x++) {
          if (CHKBIT(A->fpos, x)) {
            FOLLOWPOS(tokens, y, x) = TN(tokens, x);
          }
        }
      }
    }
    free(B);
  }

val(A) ::= val(B) QST. 
  { int y, x;
    A = newnode(NSTATE(tokens)); 
    A->na = 1;
    BMCPY(tokens, A->fpos, B->fpos);
    BMCPY(tokens, A->lpos, B->lpos); 
    free(B);
  }

val(A) ::= GRO lor(B) GRC.
  { A = B; }

val(A) ::= CHR(B). 
  { 
    A = newnode(NSTATE(tokens));
    A->na = 0;
    SETBIT(A->fpos, SN(tokens, B));
    SETBIT(A->lpos, SN(tokens, B));
  }

val(A) ::= DOT(B).
  { 
    A = newnode(NSTATE(tokens));
    A->na = 0;
    SETBIT(A->fpos, SN(tokens, B));
    SETBIT(A->lpos, SN(tokens, B));
  }
    
val(A) ::= SET(B). 
  { 
    A = newnode(NSTATE(tokens));
    A->na = 0;
    SETBIT(A->fpos, SN(tokens, B));
    SETBIT(A->lpos, SN(tokens, B));
  }

